import sys
from pathlib import Path

HERE = Path(__file__).parent.resolve()
LOCAL_APP_DIR = HERE / 'app'
LOCAL_BUILD_DIR = LOCAL_APP_DIR / 'build'

GLOBAL_QUETZ_DIR = Path(sys.prefix) / 'share' / 'quetz'
GLOBAL_FRONTEND_DIR = GLOBAL_QUETZ_DIR / 'frontend'
GLOBAL_APP_DIR = GLOBAL_FRONTEND_DIR / 'app'
GLOBAL_BUILD_DIR = GLOBAL_APP_DIR / 'build'
GLOBAL_EXTENSIONS_DIR = GLOBAL_FRONTEND_DIR / 'extensions'



import importlib
import json
import os
import shutil
import subprocess
from pathlib import Path
from shutil import which
from typing import List, Optional, Tuple

from setuptools import find_packages
from typer import Argument, Option, Typer

from .paths import (
    GLOBAL_APP_DIR,
    GLOBAL_EXTENSIONS_DIR,
    GLOBAL_FRONTEND_DIR,
    GLOBAL_QUETZ_DIR,
    HERE,
    LOCAL_APP_DIR,
)
from .utils import clean_dir, get_extensions_dir, get_federated_extensions

app = Typer()


@app.command()
def link_frontend(
    dev_mode: bool = Option(
        False, "--development", help="Whether to install it in dev mode or not"
    )
) -> None:
    """Intall the Quetz-Frontend"""
    assert LOCAL_APP_DIR.exists()

    if not GLOBAL_FRONTEND_DIR.exists():
        GLOBAL_FRONTEND_DIR.mkdir(parents=True, exist_ok=True)

    if GLOBAL_APP_DIR.exists():
        if GLOBAL_APP_DIR.is_symlink():
            GLOBAL_APP_DIR.unlink()
        else:
            shutil.rmtree(GLOBAL_APP_DIR)

    if dev_mode:
        GLOBAL_APP_DIR.symlink_to(LOCAL_APP_DIR)
        print(
            f"""Symlink created:
        Ori:  {LOCAL_APP_DIR}
        Dest: {GLOBAL_APP_DIR}
        """
        )
    else:
        shutil.copytree(LOCAL_APP_DIR, GLOBAL_APP_DIR, symlinks=True)
        print(
            f"""App directory copied:
        Ori:  {LOCAL_APP_DIR}
        Dest: {GLOBAL_APP_DIR}
        """
        )


@app.command()
def clean_frontend() -> None:
    """Clean the Quetz-Frontend"""

    if GLOBAL_APP_DIR.is_file() or GLOBAL_APP_DIR.is_symlink():
        GLOBAL_APP_DIR.unlink()
    elif GLOBAL_APP_DIR.is_dir():
        shutil.rmtree(GLOBAL_APP_DIR)


@app.command()
def install(ext_path: str = Argument(Path(), help="The path of the extension")) -> None:
    """Build and install an extension"""

    if not GLOBAL_EXTENSIONS_DIR.exists():
        os.mkdir(GLOBAL_EXTENSIONS_DIR)

    extension_path = Path(ext_path).resolve()
    assert extension_path.joinpath("package.json").exists()

    _build_extension(extension_path, True, False)

    module, metadata = _get_extensions_metadata(extension_path)
    src = Path(extension_path).joinpath(module.__name__, metadata[0]["src"])
    dest = GLOBAL_EXTENSIONS_DIR.joinpath(metadata[0]["dest"])

    clean_dir(dest)

    shutil.copytree(src, dest, symlinks=True)
    print(
        f"""
    Extension installed:
        Path:  {dest}
    """
    )


@app.command()
def develop(ext_path: str = Argument(Path(), help="The path of the extension")) -> None:
    """Build and install an extension in dev mode"""

    if not GLOBAL_EXTENSIONS_DIR.exists():
        os.mkdir(GLOBAL_EXTENSIONS_DIR)

    extension_path = Path(ext_path).resolve()
    assert extension_path.joinpath("package.json").exists()

    _build_extension(extension_path, True, False)

    _develop_extension(extension_path)


@app.command()
def build(
    ext_path: str = Argument(Path(), help="The path of the extension"),
    dev_mode: bool = Option(False, "--development", help="Build in development"),
) -> None:
    """Build an extension"""

    if not GLOBAL_EXTENSIONS_DIR.exists():
        os.mkdir(GLOBAL_EXTENSIONS_DIR)

    extension_path = Path(ext_path).resolve()
    assert extension_path.joinpath("package.json").exists()

    _build_extension(extension_path, dev_mode, False)


@app.command()
def watch(ext_path: str = Argument(Path(), help="The path of the extension")) -> None:
    """Watch an extension"""

    if not GLOBAL_EXTENSIONS_DIR.exists():
        os.mkdir(GLOBAL_EXTENSIONS_DIR)

    extension_path = Path(ext_path).resolve()
    assert extension_path.joinpath("package.json").exists()

    _develop_extension(extension_path)
    _build_extension(extension_path, True, True)


@app.command()
def uninstall(ext_name: str = Argument("", help="The name of the extension")) -> None:
    """Uninstall an extension"""

    if not GLOBAL_EXTENSIONS_DIR.exists():
        os.mkdir(GLOBAL_EXTENSIONS_DIR)

    extension_path = Path(GLOBAL_EXTENSIONS_DIR, ext_name)
    clean_dir(extension_path)


@app.command()
def list() -> None:
    """List of extensions"""

    print(f"Installed extensions:")
    print(f"---------------------")
    print(f"  Installation path: '{GLOBAL_EXTENSIONS_DIR}'\n")

    extensions, disabled_extensions = get_federated_extensions([get_extensions_dir()])

    if not extensions:
        print("No installed extensions yet")

    for ext in extensions.values():
        print(f'\t-  {Path(ext["ext_path"]).relative_to(GLOBAL_EXTENSIONS_DIR)}')

    print(f"Disabled extensions:")
    print(f"---------------------")
    for ext in disabled_extensions:
        print(f"\t-  {ext}")

    print()


@app.command()
def clean() -> None:
    """Clean the extensions directory"""
    if GLOBAL_EXTENSIONS_DIR.exists():
        shutil.rmtree(GLOBAL_EXTENSIONS_DIR)


@app.command()
def paths() -> None:
    """Quetz installation paths"""

    print(
        f"""
    System cofigured paths:
        Quetz:      {GLOBAL_QUETZ_DIR}
        Frontend:   {GLOBAL_FRONTEND_DIR}
        App:        {GLOBAL_APP_DIR}
        Extensions: {GLOBAL_EXTENSIONS_DIR}
    """
    )


def _develop_extension(ext_path: Path):
    with (ext_path / "package.json").open(encoding="utf-8") as fid:
        ext_data = json.load(fid)

    _, metadata = _get_extensions_metadata(ext_path)
    src = ext_path / ext_data["quetz"].get("outputDir", metadata[0]["src"])
    dest = GLOBAL_EXTENSIONS_DIR.joinpath(ext_data["name"])

    clean_dir(dest)

    # Create parent directory if extension name is scoped
    dest.parent.mkdir(parents=True, exist_ok=True)

    dest.symlink_to(src)
    print(
        f"""
    Symlink created:
        Ori:  {src!s}
        Dest: {dest!s}
    """
    )


def _build_extension(
    ext_path: Path,
    dev_mode: bool = False,
    watch: bool = False,
    core_path: Optional[Path] = None,
):
    if core_path is None:
        core_path = HERE / "app"
    
    if not core_path.joinpath("package.json").exists():
        print(f"Quetz frontend not found at '{core_path!s}'")

    builder_path = _find_builder(ext_path)
    if builder_path is None:
        print(f"Could not find @quetz-frontend/builder at {ext_path!s}")
        print(f"Extensions require a devDependency '@quetz-frontend/builder'")
        return

    exe = "node"
    exe_path = which(exe)

    if not exe_path:
        print(f"Could not find {exe}. Install NodeJS.")
        exit(1)

    command = [exe, str(builder_path), "--core-path", str(core_path.resolve())]

    if dev_mode:
        command.append("--development")
        command.append("--source-map")

    if watch:
        command.append("--watch")

    command.append(str(ext_path))

    print("Building extension")
    subprocess.check_call(command)


def _find_builder(ext_path: Path) -> Optional[Path]:
    """Find the package '@quetz-frontend/builder' in the extension dependencies"""

    with (ext_path / "package.json").open(encoding="utf-8") as fid:
        ext_data = json.load(fid)

    depVersion2 = ext_data.get("devDependencies", dict()).get("@quetz-frontend/builder")
    depVersion2 = depVersion2 or ext_data.get("dependencies", dict()).get(
        "@quetz-frontend/builder"
    )
    if depVersion2 is None:
        return None

    # Find @quetz-frontend/builder in the node_modules directory
    target = ext_path
    while not (target / "node_modules" / "@quetz-frontend" / "builder").exists():
        if target.parent == target:
            return None
        target = target.parent

    return (
        target
        / "node_modules"
        / "@quetz-frontend"
        / "builder"
        / "lib"
        / "build-quetzextension.js"
    )


def _get_extensions_metadata(
    module_path: Path,
) -> Tuple["importlib.ModuleType", List[str]]:
    mod_path = module_path.resolve()
    if not mod_path.exists():
        raise FileNotFoundError(f"The path `{mod_path!s}` does not exist.")

    # TODO: Change function name to match lab
    try:
        module = importlib.import_module(str(module_path))
        if hasattr(module, "js_plugin_paths"):
            return module, module.js_plugin_paths()
        else:
            module = None
    except Exception:
        module = None

    # Looking for modules in the package
    packages = find_packages(str(mod_path))
    for package in packages:
        try:
            module = importlib.import_module(package)
            if hasattr(module, "js_plugin_paths"):
                return module, module.js_plugin_paths()
        except Exception:
            module = None

    raise ModuleNotFoundError(f"There is not a extension at {module_path}")


if __name__ == "__main__":
    app()



import json
import logging
import shutil
from functools import lru_cache
from itertools import chain
from pathlib import Path
from typing import List

from .paths import GLOBAL_EXTENSIONS_DIR

logger = logging.getLogger("quetz.frontend")


def clean_dir(dir_path: Path) -> None:
    """Clean a directory"""

    if dir_path.is_file() or dir_path.is_symlink():
        dir_path.unlink()

    elif dir_path.is_dir():
        shutil.rmtree(str(dir_path))


@lru_cache(maxsize=1)
def get_extensions_dir() -> Path:
    if not GLOBAL_EXTENSIONS_DIR.exists():
        GLOBAL_EXTENSIONS_DIR.mkdir(parents=True, exist_ok=True)
        logger.info("Creating a global frontend extensions directory.")

    return GLOBAL_EXTENSIONS_DIR

def get_package_url(data):
    """Get the url from the extension data
    """
    # homepage, repository  are optional
    if 'homepage' in data:
        url = data['homepage']
    elif 'repository' in data and isinstance(data['repository'], dict):
        url = data['repository'].get('url', '')
    else:
        url = ''
    return url

def get_federated_extensions(quetzextensions_path: List[Path]) -> dict:
    """Get the metadata about federated extensions"""

    # Internal getter to apply lru_cache as it does not support list argument
    @lru_cache()
    def get_metadata(ext_dir: Path) -> list:
        datas = []

        # extensions are either top-level directories, or two-deep in @org directories
        dirs = chain(
            ext_dir.glob("[!@]*/package.json"),
            ext_dir.glob("@*/*/package.json"),
        )

        for ext_path in dirs:
            with ext_path.open(encoding="utf-8") as fid:
                pkgdata = json.load(fid)

            if pkgdata["name"] not in federated_extensions:
                data = dict(
                    name=pkgdata["name"],
                    version=pkgdata["version"],
                    description=pkgdata.get("description", ""),
                    url=get_package_url(pkgdata),
                    ext_dir=str(ext_dir),
                    ext_path=str(ext_path.parent),
                    is_local=False,
                    dependencies=pkgdata.get("dependencies", dict()),
                    quetz=pkgdata.get("quetz", dict()),
                )
                install_path = ext_path.parent / "install.json"
                if install_path.exists():
                    with install_path.open(encoding="utf-8") as fid:
                        data["install"] = json.load(fid)

                # Check if this extension disables another extension
                for disabled in data['quetz'].get('disabledExtensions', []):
                    if disabled not in disabled_extensions:
                        disabled_extensions.append(disabled)
                
                datas.append(data)

        return datas

    federated_extensions = dict()
    disabled_extensions = []
    for ext_dir in quetzextensions_path:
        datas = get_metadata(ext_dir)
        for data in datas:
            federated_extensions[data["name"]] = data

    return (federated_extensions, disabled_extensions)



"""Frontend Resources Getters."""

import json
import logging
from functools import lru_cache
from pathlib import Path
from typing import List

import jinja2

from .paths import GLOBAL_FRONTEND_DIR, LOCAL_APP_DIR


logger = logging.getLogger("quetz.frontend")


@lru_cache(maxsize=1)
def get_index_template(frontend_dir: Path) -> "jinja2.Template":
    """Get the Jinja template for index.html"""
    return jinja2.Template((frontend_dir / "static" / "index.html.j2").read_text())


@lru_cache(maxsize=1)
def get_frontend_dir() -> Path:
    """Get the frontend directory."""
    if LOCAL_APP_DIR.exists():
        logger.info("Using local DEVELOPMENT frontend directory.")
        return LOCAL_APP_DIR
    elif GLOBAL_FRONTEND_DIR.exists():
        logger.info("Using global frontend directory.")
        return GLOBAL_FRONTEND_DIR
    else:
        raise RuntimeError(
            f"Could not find frontend files in:\n- {LOCAL_APP_DIR!s}\n- {GLOBAL_FRONTEND_DIR!s}"
        )


@lru_cache(maxsize=1)
def get_frontend_settings() -> dict:
    """Get the frontend settings."""

    template_path = get_frontend_dir() / "templates"
    if template_path.exists():
        # Load settings
        with (template_path / "settings.json").open() as fi:
            data = json.load(fi)
    else:
        data = []

    return Settings(data)


class Settings:
    def __init__(self, default_settings: List[dict]):
        self.__map = dict(map(lambda s: (s["id"], s), default_settings))

    def __getitem__(self, plugin: str) -> dict:
        return self.__map[plugin]

    def __setitem__(self, plugin: str, data: dict) -> None:
        if plugin not in self.__map:
            logger.error()
            self.__map[plugin] = {"id": plugin, "raw": ""}

        self.__map[plugin]["raw"] = data

    def to_json(self) -> dict:
        return {"settings": list(self.__map.values())}




from pathlib import Path
from setuptools import setup

HERE = Path(__file__).parent.resolve()
LOCAL_APP_DIR = HERE / "quetz_frontend" / "app"
GLOBAL_APP_DIR = "share/quetz/frontend/app"

data_files_spec = [
    (GLOBAL_APP_DIR, str(LOCAL_APP_DIR), "static/**"),
    (GLOBAL_APP_DIR, str(LOCAL_APP_DIR), "themes/**"),
    (GLOBAL_APP_DIR, str(LOCAL_APP_DIR), "schemas/**")
]

# Representative files that should exist after a successful build
ensured_targets = [
    str(LOCAL_APP_DIR / "templates"),
    str(LOCAL_APP_DIR / "static"),
    str(LOCAL_APP_DIR / "themes"),
    str(LOCAL_APP_DIR / "static" / "package.json"),
    str(LOCAL_APP_DIR / "static" / "style.js"),
]

try:
    from jupyter_packaging import wrap_installers, npm_builder, get_data_files

    builder = npm_builder(build_cmd="build:prod", npm=["yarn"])
    cmdclass = wrap_installers(post_develop=builder, ensured_targets=ensured_targets)

    setup_args = {"cmdclass": cmdclass, "data_files": get_data_files(data_files_spec)}
except ImportError:
    setup_args = dict()

if __name__ == "__main__":
    setup(**setup_args)



import copy
import glob
import json
import logging
import os
import sys
from functools import lru_cache
from pathlib import Path
from typing import Optional

import jinja2
from fastapi import APIRouter, Body, Depends, Header, HTTPException
from fastapi.responses import FileResponse, HTMLResponse
from quetz import authorization, rest_models
from quetz.authentication.registry import AuthenticatorRegistry
from quetz.config import Config
from quetz.dao import Dao
from quetz.deps import get_dao, get_rules, get_session

from .utils import get_extensions_dir, get_federated_extensions
from .paths import GLOBAL_FRONTEND_DIR
from .resources import (
    Settings,
    get_frontend_dir,
    get_frontend_settings,
    get_index_template,
)

logger = logging.getLogger("quetz.frontend")
config = Config()

router = APIRouter()
catchall_router = APIRouter()

ROUTE_PREFIX = "ui"

config_data = {}
federated_extensions = []


@lru_cache(maxsize=1)
def is_dev_mode() -> bool:
    """Whether the server run in dev mode or not.

    The application is considered in dev mode if the
    ``--reload`` cli option is set.
    """
    return "--reload" in sys.argv


# Endpoint handlers


@router.get("/api/settings", include_in_schema=False)
def get_settings(frontend_settings: Settings = Depends(get_frontend_settings)):
    return frontend_settings.to_json()


@router.get("/api/settings/{extension:path}", include_in_schema=False)
def get_settings(
    extension: str, frontend_settings: Settings = Depends(get_frontend_settings)
):
    return frontend_settings[extension]


@router.put("/api/settings/{extension:path}", include_in_schema=False, status_code=204)
def put_settings(
    extension: str,
    settings: str = Body(...),
    frontend_settings: Settings = Depends(get_frontend_settings),
):
    logger.warn(settings)
    # Override the settings
    frontend_settings[extension] = json.loads(settings)["raw"]
    return {"success": True}


@router.get("/themes/{resource:path}", include_in_schema=False)
def get_theme(
    resource: str,
    frontend_dir: Path = Depends(get_frontend_dir),
    extensions_dir: Path = get_extensions_dir(),
):
    logger.warn(resource)
    path = frontend_dir / "themes" / resource
    if path.exists():
        return FileResponse(path=path)
    else:
        logger.warn(str(extensions_dir))
        # Path.glob does not follow symlinks
        for theme_extension in glob.glob(str(extensions_dir) + "/**/themes", recursive=True):
            # TODO Replace local paths with mangled paths; e.g. `url('../foo.css')`, `url('images/foo.png')`
            path = Path(theme_extension) / resource
            logger.warn(str(path))
            if path.exists():
                return FileResponse(path)

    raise HTTPException(status_code=404)


@router.get("/extensions/{resource:path}", include_in_schema=False)
def extensions(
    resource: str,
    session: dict = Depends(get_session),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    # frontend_dir: Path = Depends(get_frontend_dir),
    extensions_dir: Path = Depends(get_extensions_dir),
):
    path = extensions_dir / resource
    if path.exists():
        return FileResponse(path=path)
    else:
        raise HTTPException(status_code=404)


@router.get("/static/{resource:path}", include_in_schema=False)
def static(
    resource: str,
    session: dict = Depends(get_session),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    frontend_dir: Path = Depends(get_frontend_dir),
):
    path = frontend_dir / "static" / resource
    if path.exists():
        return FileResponse(
            path=path, headers={"cache-control": "public, max-age=604800, immutable"}
        )
    else:
        raise HTTPException(status_code=404)


@catchall_router.get("/{resource:path}", include_in_schema=False)
def index(
    resource: str,
    session: dict = Depends(get_session),
    dao: Dao = Depends(get_dao),
    auth: authorization.Rules = Depends(get_rules),
    frontend_dir: Path = Depends(get_frontend_dir),
    extensions_dir: Path = Depends(get_extensions_dir),
    cache_control: Optional[str] = Header(None),
):
    global config_data
    user_id = auth.get_user()
    profile = dao.get_profile(user_id)

    if "." in resource:
        file_name = (
            resource
            if ("icons" in resource or "logos" in resource or "page-data" in resource)
            else resource.split("/")[-1]
        )

        path = frontend_dir / "static" / file_name
        if path.exists():
            return FileResponse(path=path)
        else:
            raise HTTPException(status_code=404)
    else:
        extensions, disabled_extensions = get_federated_extensions([extensions_dir])
        federated_extensions = load_federated_extensions(extensions)
        config_data["federated_extensions"] = federated_extensions
        config_data["disabledExtensions"] = disabled_extensions

        logger.info(cache_control)
        if is_dev_mode():
            get_index_template.cache_clear()

        index_template = get_index_template(frontend_dir)
        index_rendered = get_rendered_index(config_data, profile, index_template)
        return HTMLResponse(content=index_rendered, status_code=200)


def under_frontend_dir(frontend_dir: Path, path: Path) -> bool:
    """
    Check that path is under frontend_dir
    """
    path = path.resolve()
    fdir = frontend_dir.resolve()
    globalDir = GLOBAL_FRONTEND_DIR.resolve()
    return (
        Path(os.path.commonpath([path, fdir])) == fdir
        or Path(os.path.commonpath([path, globalDir])) == globalDir
    )


def get_rendered_index(
    config_data: dict, profile: Optional["Profile"], index_template: "jinja2.Template"
) -> str:
    """Rendered the index file with user profile"""
    cfg = copy.copy(config_data)
    if profile is not None:
        cfg["logged_in_user_profile"] = rest_models.Profile.from_orm(profile).json()
    elif "logged_in_user_profile" in cfg:
        del cfg["logged_in_user_profile"]

    index_rendered = index_template.render(page_config=cfg)
    return index_rendered


def load_federated_extensions(federated_extensions: dict) -> list:
    """Load the list of extensions"""
    extensions = []
    for name, data in federated_extensions.items():
        build_info = data["quetz"]["_build"]
        build_info["name"] = name
        extensions.append(build_info)

    return extensions


def register(app):
    global config_data

    app.include_router(router, prefix=f"/{ROUTE_PREFIX}")
    app.include_router(catchall_router)

    # frontend_dir = config.general_frontend_dir
    # extensions_dir = config.general_extensions_dir
    frontend_dir = get_frontend_dir()
    extensions_dir = get_extensions_dir()

    logger.info(f"Configured frontend found: {frontend_dir!s}")
    logger.info(f"Configured extensions directory: {extensions_dir!s}")

    extensions, disabled_extensions = get_federated_extensions([extensions_dir])
    federated_extensions = load_federated_extensions(extensions)

    auth_registry = AuthenticatorRegistry()
    google_login_available = auth_registry.is_registered("google")
    github_login_available = auth_registry.is_registered("github")
    gitlab_login_available = auth_registry.is_registered("gitlab")
    azuread_login_available = auth_registry.is_registered("azuread")

    config_data = {
        "appName": "Quetz – the fast conda package server!",
        "baseUrl": "/",
        "wsUrl": "",
        "appUrl": f"/{ROUTE_PREFIX}",
        "labextensionsUrl": f"/{ROUTE_PREFIX}/extensions",
        "themesUrl": f"/{ROUTE_PREFIX}/themes",
        "settingsUrl": f"/{ROUTE_PREFIX}/api/settings",
        "listingsUrl": "",
        "fullAppUrl": f"/{ROUTE_PREFIX}",
        "fullStaticUrl": f"/{ROUTE_PREFIX}/static",
        "fullQuetzextensionsUrl": f"/{ROUTE_PREFIX}/extensions",
        "fullThemesUrl": f"/{ROUTE_PREFIX}/themes",
        "fullSettingsUrl": f"/{ROUTE_PREFIX}/api/settings",
        "fullListingsUrl": "",
        "federated_extensions": federated_extensions,
        "disabledExtensions": disabled_extensions,
        "github_login_available": github_login_available,
        "gitlab_login_available": gitlab_login_available,
        "google_login_available": google_login_available,
        "azuread_login_available": azuread_login_available,
        "cacheFiles": False,
        "devMode": False,
        "mode": "multiple-document",
        "exposeAppInBrowser": False,
        "cacheFiles": False,
        "devMode": False,
        "mode": "multiple-document",
        "exposeAppInBrowser": False,
    }
